#include <stdio.h>
#include <string.h>
#include <curses.h>
#include <math.h>
#include <assert.h>
//#include "main.h"       // our own stuff


  //
  //             R-20 SIMULATOR
  //
  //      jul/aug 2012, insektoiden.de


  // "curses" stuff
  WINDOW *wnd;
  int nrows;
  int ncols;


  // data representation

  // How to represent the 4 registers?
  // Originally, there were 12 nibbles for BCD digits, 2 nibbles for
  // BCD coded exponent and 1 nibble each for the signs.
  //
  //   +-   +-    1 ,  2    3    4    5    6    7    8    9    0    1    2
  // |....|....|....|....|....|....|....|....|....|....|....|....|....|....|
  //
  //  sign sign  1G  100M  10M   1M 100k  10k  1k   100  10    1   10    1
  //  man  exp  \_________________ man _________________________/ \__exp__/
  //
  // | Byte 0  |  1 |  2 |  3 |  4 |  5 |  6 |  7 |  8 |  9 | 10 | 11 | 12 |
  //
  //
  // sign     [0001] or [1111],  two nibbles       Byte   0
  // mantissa [0..9] ... [0..9], ten bcd digits    Bytes 1-5
  // exponent [0..9] + [0..9],   two bcd digits    Byte   6

  // New coding as of 2012-08-03:
  //
  // Each digit gets it dedicated byte to ease addition etc.
  // Signs are stored in one byte:  0b .... ..me  (mantissa, exponent)

#define FLAGS        0
#define M            1
#define M_HIGH       1
#define M_ONES      10
#define E           11
#define E_TENS      11
#define E_ONES      12
#define NUMBER_SIZE 13

#define NEG_M     0x01
#define NEG_E     0x02
#define INFTY     0x04

#define X      0
#define Y      1
#define Z      2
#define T      3
#define A      4
#define B      5
#define REG_NR 6

#define DEBUG_LINE 12     // where to print debugging output

  char reg[REG_NR][NUMBER_SIZE];




/**
 * Convert register to string.
 */
 // "--" "1.234567890" "12" --> "-1.234567890e-12"
 //
void rtostr(char* str, char* r) {

  // check for infinity
  if (r[FLAGS] & INFTY) {
    if (r[FLAGS] & NEG_M) strcpy(str, " neg infinity!  ");
    else strcpy(str, "   infinity!    ");
    return;
  }

  if (r[FLAGS] & NEG_M) str[0] = '-';     // check for man negative
  else str[0] = ' ';

  { assert(r[M_HIGH]>=0 && r[M_HIGH]<=9); }
  str[1] = r[M_HIGH] + '0';    // access first digit of mantissa
  str[2] = '.';
  int i;                  // access digit 2-10 of mantissa
  for (i=1; i<=9; i++) str[i+2] = r[M_HIGH+i] + '0';
  str[12] = 'e';

  if (r[FLAGS] & NEG_E) str[13] = '-';    // check for exp negative
  else str[13] = ' ';

  str[14] = r[E_TENS] + '0';   // access digits of exponent
  str[15] = r[E_ONES] + '0';
  str[16] = '\0';
}

/**
 * Print values of t,z,y,x registers.
 */
void screen_refresh() {
  char str[17];   // will hold number, e.g. "-1.234567890e-12\0"
  memset(str, ' ', 16);
  str[16] = '\0';

  move(0,0);      // also possible: clear();

  printw("+------------------+\n");
  printw("| ");
  rtostr(str, reg[T]);
  printw(str);
  printw(" |\n");
  printw("| ");
  rtostr(str, reg[Z]);
  printw(str);
  printw(" |\n");
  printw("| ");
  rtostr(str, reg[Y]);
  printw(str);
  printw(" |\n");
  printw("| ");
  rtostr(str, reg[X]);
  printw(str);
  printw(" |\n");
  printw("+------------------+");

  move(nrows-1, ncols-1);       // "hide" cursor
} // screen_refresh


void err_error(char* str) {
  move(4, 0);
  printw(str);
  move(nrows-1, ncols-1);       // "hide" cursor
  getch();
}

/**
 * Copy contents of src to dst, whose contents will be lost.
 */
void copy_reg(char* dst, char* src) {
  memcpy(dst, src, NUMBER_SIZE);
}

/**
 * Return exponent of register.
 */
int get_e(char* r) {
  int e;

  e = r[E_TENS]*10 + r[E_ONES];
  if (r[FLAGS] & NEG_E) e = -e;
  return e;
}


/**
 * Debugging helper function to print one line outside the main screen.
 */
void debug_print(char* str) {
  move(10,0);
  printw("                                         ");
  move(10,0);
  printw(str);
}


/**
 * Increase exponent by 1.
 */
void inc_e(char* r) {

  if (r[FLAGS] & NEG_E) {                       // exp sign -
    if ((r[E_TENS] == 0) && (r[E_ONES] == 1)) { // exp = -1
      r[FLAGS] ^= NEG_E;                        // flip sign, is + now
      r[E_TENS] = 0;                            // exp = +0
      r[E_ONES] = 0;
    } else {                                    // exp < -1
      r[E_ONES]--;                              // just decreasing
      if (r[E_ONES] == -1) {                    // but watch for underflows
        r[E_TENS]--;
        r[E_ONES] = 9;
      }
    }
  } else {                                      // exp sign +
    if ((r[E_TENS] == 9) && (r[E_ONES] == 9)) { // exp = +99
      memset(r[1], 0, NUMBER_SIZE-1);           // clear r, save FLAGS
      r[FLAGS] |= INFTY;                        // set r to INFTY
    } else {                                    // exp < +99
      r[E_ONES]++;                              // just increasing
      if (r[E_ONES] == 10) {                    // but watch for overflows
        r[E_ONES] = 0;
        r[E_TENS]++;
      }
    }
  }
}

/**
 * Shift mantissa 1 digit to the right.
 *
 * Last digit is lost, but correct rounding on second-to-last digit
 * is assured.
 */
void shiftr1(char* r) {
  inc_e(r);
  if (r[M_ONES] >= 5) r[M_ONES-1]++;      // correct rounding
  int i;
  for (i=M_ONES; i>=M_HIGH; i--) r[i] = r[i-1];
  r[M_HIGH] = 0;
}

/**
 * Shift mantissa n digits to the right.
 */
void shiftr(char* r, int n) {
  { assert(n > 0); }
  for (; n > 0; n--) shiftr1(r);
}


/**
 * Push 3.141592654 (@f$ \approx \pi @f$) on stack.
 */
key_pi() {
  copy_reg(reg[T], reg[Z]);
  copy_reg(reg[Z], reg[Y]);
  copy_reg(reg[Y], reg[X]);

  reg[X][FLAGS] = 0x00;
  reg[X][M_HIGH]   = 3;
  reg[X][M+1] = 1;
  reg[X][M+2] = 4;
  reg[X][M+3] = 1;
  reg[X][M+4] = 5;
  reg[X][M+5] = 9;
  reg[X][M+6] = 2;
  reg[X][M+7] = 6;
  reg[X][M+8] = 5;
  reg[X][M_ONES] = 4;
  reg[X][E_TENS]   = 0;
  reg[X][E_ONES] = 0;
}


/**
 * Add two registers and store result in third one.
 *
 * res <-- r1 + r2.
 * Even in case of overflow, res is changed in the process.
 */
void plus(char* res, char* r1, char* r2) {
  memset(res, 0, NUMBER_SIZE);            // result is cleared first
  int e1;
  int e2;

  e1 = get_e(r1);
  e2 = get_e(r2);

  if (e1 > e2) shiftr(r2, e1-e2);
  else if (e2 > e1) shiftr(r1, e2-e1);

  // Ok, exponents equalized. Now do actual addition of mantissa.

  res[E_TENS] = r1[E_TENS];   // use same exponent
  res[E_ONES] = r1[E_ONES];

  if ((!r1[FLAGS] & NEG_M) && (!r2[FLAGS] & NEG_M)) {   // both positive
    int i;
    for (i=M_ONES; i>M_HIGH; i--) {
      res[i] += r1[i] + r2[i];
      if (res[i] >= 10) {                 // overflow
        res[i-1]++;
        res[i] -= 10;
      }
    }
    res[M_HIGH] += r1[M_HIGH] + r2[M_HIGH];
    if (res[M_HIGH] >= 10) {
      res[M_HIGH] -= 10;
      shiftr1(res);
      inc_e(res);
      if (res[FLAGS] & INFTY) {
        overflow(res);
        return;
      }
    }
  } else
  if ((r1[FLAGS] & NEG_M) && (r2[FLAGS] & NEG_M)) {  // both negative
    r1[FLAGS] ^= NEG_M;   // r1 = -r1
    r2[FLAGS] ^= NEG_M;   // r2 = -r2
    plus(res, r1, r2);                // TODO watch out for overflow here!
    res[FLAGS] ^= NEG_M;  // res = -res
  } else
  if ((r1[FLAGS] & NEG_M) ^ (r2[FLAGS] & NEG_M)) {   // one neg, one pos
    if (r1[FLAGS] & NEG_M) {
      r1[FLAGS] ^= NEG_M;   // r1 = -r1
      r2[FLAGS] ^= NEG_M;   // r2 = -r2
    }
//    sub(res, r1, r2);
  }
}


/**
 * Deletes top of stack.
 */
void pop() {
  copy_reg(reg[X], reg[Y]);
  copy_reg(reg[Y], reg[Z]);
  copy_reg(reg[Z], reg[T]);
}


/**
 * Put argument into X register and shifts the stack.
 */
void push(char* r) {
  copy_reg(reg[T], reg[Z]);
  copy_reg(reg[Z], reg[Y]);
  copy_reg(reg[Y], reg[X]);
  copy_reg(reg[X], r);
}


/**
 * Print message according to status of argument.
 */
void overflow(char* r) {
  if (r[FLAGS] & INFTY & NEG_M) err_error(" invalid input  ");
  else if (r[FLAGS] & INFTY) err_error(" invalid input  ");
  else err_error(" general overflw");
}


/**
 * Called when "+" is pressed.
 */
key_plus() {
  plus(reg[A], reg[X], reg[Y]);
  if (reg[A][FLAGS] & INFTY) overflow(reg[A]);
  else {
    pop();
    pop();
    push(reg[A]);
  }
screen_refresh();
getch();
exit_curses();
}


/**
 * Perform stack function "enter".
 */
enter() {
  copy_reg(reg[T], reg[Z]);
  copy_reg(reg[Z], reg[Y]);
  copy_reg(reg[Y], reg[X]);
}


/**
 * Called when "enter" is pressed.
 */
key_enter() {
  enter();
}

/**
 * Call subroutine assigned to pressed key.
 */
void call(int key) {
  switch(key) {
//    case 's': { key_sin(); break; }
//    case 'a': { key_asin(); break; }
    case 'p': { key_pi(); break; }
//    case '0': { key_digit(0); break; }
//    case '1': { key_digit(1); break; }
//    case '2': { key_digit(2); break; }
//    case '3': { key_digit(3); break; }
//    case '4': { key_digit(4); break; }
//    case '5': { key_digit(5); break; }
//    case '6': { key_digit(6); break; }
//    case '7': { key_digit(7); break; }
//    case '8': { key_digit(8); break; }
//    case '9': { key_digit(9); break; }
    case 10: { key_enter(); break; }
    case '+': { key_plus(); break; }
  }
}


void init_regs() {
  //example
  reg[T][FLAGS]  = NEG_M | NEG_E;
  reg[T][M_HIGH] = 3;
  reg[T][M+1]    = 1;
  reg[T][M+2]    = 4;
  reg[T][M+3]    = 1;
  reg[T][M+4]    = 5;
  reg[T][M+5]    = 9;
  reg[T][M+6]    = 2;
  reg[T][M+7]    = 6;
  reg[T][M+8]    = 5;
  reg[T][M_ONES] = 4;
  reg[T][E_TENS] = 0;
  reg[T][E_ONES] = 0;

  reg[Z][FLAGS]  = NEG_M;
  reg[Z][M_HIGH] = 5;
  reg[Z][M+1]    = 2;
  reg[Z][M+2]    = 6;
  reg[Z][M+3]    = 7;
  reg[Z][M+4]    = 3;
  reg[Z][M+5]    = 7;
  reg[Z][M+6]    = 2;
  reg[Z][M+7]    = 3;
  reg[Z][M+8]    = 6;
  reg[Z][M_ONES] = 7;
  reg[Z][E_TENS] = 2;
  reg[Z][E_ONES] = 0;

  reg[Y][FLAGS]  = 0x00;
  reg[Y][M_HIGH] = 9;
  reg[Y][M+1]    = 8;
  reg[Y][M+2]    = 2;
  reg[Y][M+3]    = 3;
  reg[Y][M+4]    = 4;
  reg[Y][M+5]    = 6;
  reg[Y][M+6]    = 3;
  reg[Y][M+7]    = 7;
  reg[Y][M+8]    = 5;
  reg[Y][M_ONES] = 3;
  reg[Y][E_TENS] = 2;
  reg[Y][E_ONES] = 0;

  reg[X][FLAGS]  = 0x00;
  reg[X][M_HIGH] = 7;
  reg[X][M+1]    = 4;
  reg[X][M+2]    = 4;
  reg[X][M+3]    = 7;
  reg[X][M+4]    = 3;
  reg[X][M+5]    = 3;
  reg[X][M+6]    = 7;
  reg[X][M+7]    = 3;
  reg[X][M+8]    = 4;
  reg[X][M_ONES] = 7;
  reg[X][E_TENS] = 2;
  reg[X][E_ONES] = 0;
} // init_regs


/**
 * Debugging function: print key code below output.
 */
void print_key(int key) {
  move(DEBUG_LINE,0);
  printw("              ");
  move(DEBUG_LINE,0);
  printw("Taste: %i", key);
}


/**
 * Used to kill program during debugging.
 */
void exit_curses() {
  echo();
  endwin();
  exit(0);
}


/**
 * Main loop: Output and wait for keypress.
 */
int main(void) {
  wnd = initscr();                // initialize curses
  cbreak();                       // send keys directly to program
  noecho();                       // don't write my typing to screen
  getmaxyx(wnd, nrows, ncols);    // check terminal dimensions

  init_regs();
  int key = 0;
  do {
    screen_refresh();
    key = getch();
    //print_key(key);
    call(key);
  } while (key != 27);            // 27 = escape (ASCII)

  echo();
  endwin();
  return 0;
} // main
